from uuid import uuid4

from django.contrib.postgres.indexes import BrinIndex
from django.db import models
from django.utils.translation import gettext_lazy as _


class BaseModel(models.Model):
    """ BaseModel should be added to every model """

    uuid = models.UUIDField(_("Идентификатор"), unique=True, default=uuid4, editable=False)
    is_deleted = models.BooleanField(_("Удалено ли?"), default=False)

    created_at = models.DateTimeField(_("дата создания"), auto_now_add=True)
    created_by_id = models.UUIDField(_("Создатель"), null=True)

    updated_at = models.DateTimeField(_("дата изменения"), auto_now=True)
    updated_by_id = models.UUIDField(_("Кто изменил"), null=True)

    ordering_fields = (
        'uuid',
        'created',
        'modified',
    )

    searched_fields = (
        'uuid',
        'created',
        'modified',
    )

    queryable_fields = ("uuid", "created_at", "updated_at", "created_by_id", "updated_by_id")
    orderable_fields = ("uuid", "created_at", "updated_at", "created_by_id", "updated_by_id")
    searchable_fields = ("uuid", "created_at", "updated_at", "created_by_id", "updated_by_id")
    searched_values = ("uuid", "created_at", "updated_at", "created_by_id", "updated_by_id")

    class Meta:
        abstract = True
        indexes = (
            models.Index(fields=('pk',)),
            BrinIndex(fields=('uuid',)),
            models.Index(fields=('is_deleted',)),
        )



from typing import Optional, Any


def search_params(
    search: Optional[str] = None,
    query: str = None,
    page: Optional[Any] = None,
    ordering: Optional[str] = None,
):
    return {
        "search": search,
        "query": query,
        "page": page,
        "ordering": ordering,
    }


__all__ = ['search_params']



from functools import wraps


def only_decorator(func: callable):

    @wraps(func)
    def only_wrapper(objects, *args, only=(), **kwargs):
        return func(objects, *args, **kwargs).only(*only)

    return only_wrapper


def get_related_decorator(func: callable):

    @wraps(func)
    def get_related_wrapper(objects, *args, select_related=(), prefetch_related=(), **kwargs):
        return func(objects, *args, **kwargs).select_related(*select_related).prefetch_related(*prefetch_related)

    return get_related_wrapper


def first_object_decorator(func: callable):

    @wraps(func)
    def first_object_wrapper(objects, *args, first=False, **kwargs):
        service_result = func(objects, *args, **kwargs)

        if first:
            return service_result.first()

        return service_result

    return first_object_wrapper


def ordering_decorator(func: callable):

    @wraps(func)
    def ordering_wrapper(objects, *args, ordering=None, **kwargs):
        service_result = func(objects, *args, **kwargs)

        if ordering is None:
            return service_result

        return service_result.order_by(*ordering)

    return ordering_wrapper


def pagination_decorator(func: callable):

    @wraps(func)
    def pagination_wrapper(objects, *args, limit=None, offset=None, **kwargs):
        service_result = func(objects, *args, **kwargs)

        if offset is not None:
            service_result = service_result[offset:]
        if limit is not None:
            service_result = service_result[:limit]

        return service_result

    return pagination_wrapper


__all__ = [
    'only_decorator',
    'get_related_decorator',
    'first_object_decorator',
    'ordering_decorator',
    'pagination_decorator'
]



import json
import re
from enum import Enum
from abc import abstractmethod, ABC
from typing import Optional, Any

from velait.common.services.exceptions import SearchError


class SearchOperator(Enum):
    LESS = "less"
    EQUAL = "equal"
    GREATER = "greater"
    CONTAINS = "contains"
    LESS_OR_EQUAL = "lessOrEqual"
    GREATER_OR_EQUAL = "greaterOrEqual"


class Search(ABC):
    def _parse_json(self, data, key_name: str):
        try:
            return json.loads(data)
        except (json.JSONDecodeError, TypeError, ValueError):
            if data is not None:
                raise SearchError(
                    name=key_name,
                    description=f"'{key_name}' cannot be parsed as JSON",
                )

    def __init__(self, search: Optional[str], query: str, ordering: Optional[str], model):
        self.model = model
        self._search = search
        self._query = self._parse_json(data=query, key_name='query') or []
        self._ordering = tuple(map(str.strip, ordering.split(","))) if ordering else ()
        self.validate()

    def _validate_query_part(self, query_part: dict):
        name = query_part.get('fn')
        operation = query_part.get('op')

        if (name is None) or (operation is None):
            raise SearchError(
                name='query',
                description="All items in 'query' must have 'fn', 'op', 'fv' keys",
            )
        elif name not in self.model.queryable_fields:
            raise SearchError(
                name="query",
                description=f"{name} was not found as a value",
            )

    def _format_query_part(self, query_part: dict) -> dict:
        return {
            **query_part,
            "fn": re.sub(r'(?<!^)(?=[A-Z])', '_', query_part["fn"]).lower(),
        }

    def validate(self):
        if isinstance(self._query, list):
            for i in range(len(self._query)):
                self._query[i] = self._format_query_part(self._query[i])
                self._validate_query_part(self._query[i])
        elif self._query is not None:
            raise SearchError(
                name="query",
                description=f"Field 'query' must be a list of filters",
            )

        ordering = []

        for field in self._ordering:
            if field.startswith('-'):
                ordering.append(field[1:])
            else:
                ordering.append(field)

        not_orderable_fields = set(ordering) - set(self.model.orderable_fields)

        if self._ordering and not_orderable_fields:
            parsed_fields = ' ,'.join(f"'{field}'" for field in not_orderable_fields)
            raise SearchError(
                name="ordering",
                description=f"Fields {parsed_fields} cannot be used in 'ordering'",
            )

    @abstractmethod
    def _parse_operator(self, field_name: str, operator: str, field_value: str):
        raise NotImplementedError("_parse_operator() is not implemented")

    def _parse_ordering(self):
        return self._ordering

    @abstractmethod
    def search(self):
        raise NotImplementedError("search() is not implemented")

    def parse_query_filters(self):
        return [self._parse_operator(
            field_name=query_part.get('fn'),
            operator=query_part.get('op'),
            field_value=query_part.get('fv'),
        ) for query_part in self._query]


__all__ = [
    'Search',
    'SearchError',
    'SearchOperator'
]



from velait.velait_fastapi.api.pagination.schemas import Page
from velait.velait_fastapi.api.pagination.pagination import paginate
from velait.common.services.pagination import get_page_limits, get_offset_limits



from typing import Iterable

from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

from velait.velait_fastapi.api.users.services import SSOService


class UserMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        sso_service: SSOService,
        excluded_routes: Iterable = (),
        excluded_methods: Iterable[str] = (),
        *args,
        **kwargs,
    ):
        super(UserMiddleware, self).__init__(*args, **kwargs)
        self.sso_service = sso_service
        self.excluded_routes = excluded_routes
        self.excluded_methods = excluded_methods

    async def dispatch(self, request: Request, call_next):
        is_excluded_route = any(request.url.path.startswith(route) for route in self.excluded_routes)

        if is_excluded_route or request.method in self.excluded_methods:
            return await call_next(request)

        request.scope['user'] = await self.sso_service.get_user_info(
            request.headers.get('Authorization'))
        return await call_next(request)


__all__ = ['UserMiddleware']



import json
from typing import Iterable

from fastapi import Request
from starlette.datastructures import URL

from velait.velait_fastapi.api.pagination.schemas import Page, PageInfo


def __create_param(page: int, size: int):
    return json.dumps({"page": page, "size": size})


def get_next_page(url: URL, current_page: int, last_page: int, page_size: int):
    if current_page == last_page:
        return None

    return str(url.include_query_params(page=__create_param(page=last_page, size=page_size)))


def get_prev_page(url: URL, current_page: int, page_size: int):
    if current_page == 0:
        url.remove_query_params('page')
        return str(url)

    return str(url.include_query_params(page=__create_param(page=current_page - 1, size=page_size)))


def paginate(
    request: Request,
    page_size: int,
    data: Iterable,
    total_count: int,
) -> Page:
    last_page = total_count // page_size
    current_page = request.path_params.get('page', 0)

    return Page(
        results=list(data),
        pagination=PageInfo(
            total_records=total_count,
            total_pages=1 if total_count < page_size else last_page,
            first=str(request.url.remove_query_params('page')),
            last=str(request.url.include_query_params(page=__create_param(page=last_page, size=page_size))),
            next=get_next_page(
                url=request.url,
                last_page=last_page,
                current_page=current_page,
                page_size=page_size,
            ),
            previous=get_prev_page(
                url=request.url,
                current_page=current_page,
                page_size=page_size,
            ),
        ),
    )


__all__ = [
    'paginate',
]



from uuid import UUID

from django.db.models import QuerySet
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from velait.common.exceptions import AlreadyDeletedError
from velait.velait_django.main.models import BaseModel
from velait.velait_django.main.services.decorators import (
    only_decorator,
    get_related_decorator,
    ordering_decorator,
    pagination_decorator,
    first_object_decorator,
)


User = get_user_model()


def get_object(objects, *args, **kwargs):
    return filter_objects(objects, *args, **kwargs, first=True)


@first_object_decorator
@only_decorator
@pagination_decorator
@ordering_decorator
@get_related_decorator
def filter_objects(objects, *args, **kwargs) -> QuerySet:
    return objects.filter(*args, **kwargs)


@first_object_decorator
@only_decorator
@pagination_decorator
@ordering_decorator
@get_related_decorator
def all_objects(objects):
    return filter_objects(objects)


def empty_objects(objects) -> QuerySet:
    return objects.none()


def object_exists(objects, **kwargs) -> bool:
    return filter_objects(objects, **kwargs).exists()


def delete_object(model: BaseModel):
    if model.is_deleted:
        raise AlreadyDeletedError(name="obj", description=_("Уже удалено"))

    model.is_deleted = True
    model.save(update_fields=['is_deleted'])


def update_object(model: BaseModel, updated_by_id: UUID, **kwargs):
    for key, value in kwargs.items():
        setattr(model, key, value)

    model.updated_by_id = updated_by_id
    model.save(update_fields=('updated_by_id', *kwargs.keys()))


__all__ = [
    'get_object',
    'filter_objects',
    'all_objects',
    'empty_objects',
    'object_exists',
    'update_object',
    'delete_object',
]



from djangorestframework_camel_case.render import CamelCaseJSONRenderer

from velait.velait_django.main.api.responses import APIResponse


class VelaitAPIRenderer(CamelCaseJSONRenderer):
    def render(self, data, *args, **kwargs):
        if isinstance(data, dict):
            if data.get('pagination') is None:
                data = APIResponse.format_data(data)

        return super(VelaitAPIRenderer, self).render(data, *args, **kwargs)


__all__ = ['VelaitAPIRenderer']



from velait.velait_django.velait_users.views import (
    user_logout_view,
    user_login_view,
    user_authentication_view,
)
from django.urls import path

app_name = "apps.velait_users"

urlpatterns = [
    path("logout/", user_logout_view, name="logout"),
    path("login/", user_login_view, name="login"),
    path('authenticate/', user_authentication_view, name="authenticate"),
]



from typing import Optional, Any

from fastapi import Query
from assimilator.core.database import Repository

from velait.common.services.exceptions import PaginationLimitsError
from velait.common.services.pagination import get_page_limits
from velait.common.services.search import Search, SearchError, SearchOperator


class AlchemyRepositorySearch(Search):
    def __init__(
        self,
        repository: Repository,
        search_: Optional[str] = None,
        query: str = Query(default=None),
        ordering: Optional[str] = None,
        page: Optional[Any] = None,
        default_pagination_size: int = 10,
    ):
        self.repository = repository
        self._page = self._parse_json(page, key_name='page') if (page is not None) else None
        self.default_pagination_size = default_pagination_size
        super(AlchemyRepositorySearch, self).__init__(
            search=search_,
            query=query,
            ordering=ordering,
            model=repository.model,
        )

    def validate(self):
        super(AlchemyRepositorySearch, self).validate()

        if self._page is None:
            self._page = {
                "offset": 1,
                "size": self.default_pagination_size,
            }
            return

        if not isinstance(self._page, dict):
            raise SearchError(
                name="page",
                description="Invalid structure of 'page'",
            )

        if self._page.get('offset', 0) <= 0:
            self._page['offset'] = 1

        if self._page.get('size', 0) <= 0:
            self._page['size'] = self.default_pagination_size

    def _parse_ordering(self): return self._ordering

    def search(self):
        try:
            filter_spec = self.repository.specifications.filter(*self.parse_query_filters())

            offset, limit = get_page_limits(
                page=self._page['offset'] - 1,
                page_size=self._page['size'],
            )
            pagination_spec = self.repository.specifications.paginate(offset=offset, limit=limit)

            if self._ordering:
                return self.repository.filter(
                    filter_spec,
                    pagination_spec,
                    self.repository.specifications.order(*self._parse_ordering()),
                    lazy=True,
                )

            return self.repository.filter(filter_spec, pagination_spec, lazy=True)

        except PaginationLimitsError:
            raise SearchError(
                name='page',
                description="'page' parameter contains invalid values",
            )
        except Exception:
            raise SearchError(
                name="search",
                description="Search could not be conducted",
            )

    def _parse_operator(self, field_name: str, operator: str, field_value: str):
        """
        Creates an expression object if all input operators are valid.
        If they are not, raises op an exception
        """

        if operator == "equals":
            return getattr(self.repository.model, field_name) == field_value
        elif operator == "lessOrEqual":
            return getattr(self.repository.model, field_name) <= field_value
        elif operator == "greaterOrEqual":
            return getattr(self.repository.model, field_name) >= field_value
        elif operator == "greater":
            return getattr(self.repository.model, field_name) > field_value
        elif operator == "less":
            return getattr(self.repository.model, field_name) < field_value
        elif operator == "contains":
            return getattr(self.repository.model, field_name).in_(field_value)
        else:
            raise SearchError(
                name="query",
                description=f"Operation '{operator}' is unknown",
            )


__all__ = [
    'SearchError',
    'SearchOperator',
    'AlchemyRepositorySearch',
]



from typing import TypeVar, Type, Optional

from assimilator.alchemy.database import AlchemyRepository as BaseAlchemyRepository
from assimilator.core.database import SpecificationType

from velait.common.exceptions import AlreadyDeletedError


class VelaitRepository(BaseAlchemyRepository):
    def filter(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query=None,
    ):
        return super(VelaitRepository, self).filter(
            *specifications,
            lazy=lazy,
            initial_query=initial_query,
        )

    def get(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query=None,
    ):
        return super(VelaitRepository, self).get(
            *specifications,
            lazy=lazy,
            initial_query=initial_query,
        )

    def delete(
        self,
        obj=None,
        *specifications: SpecificationType
    ) -> None:
        if obj is not None:
            if obj.is_deleted:
                raise AlreadyDeletedError()

            obj.is_deleted = True
        else:
            self.update(*specifications, is_deleted=True)


__all__ = [
    'VelaitRepository',
]



from typing import List
from uuid import UUID

from velait.velait_fastapi.api.schemas import ReadBaseSchema


class UserSchema(ReadBaseSchema):
    id: UUID
    username: str
    email: str
    permissions: List[str]


__all__ = ['UserSchema']



import re
from typing import Type

from django.conf import settings
from django.db.models import Q
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework.generics import GenericAPIView
from rest_framework.request import Request
from django.utils.translation import gettext_lazy as _
from rest_framework.serializers import ModelSerializer

from velait.common.exceptions import VelaitError
from velait.velait_django.main.models import BaseModel
from velait.velait_django.main.api.serializers import BaseSerializer
from velait.velait_django.main.api.responses import APIResponse
from velait.velait_django.main.api.pagination import VelaitPagination, VelaitPaginationInspector
from velait.velait_django.main.services.services import filter_objects
from velait.velait_django.main.services.search import SearchError, DjangoSearch


class SearchView(GenericAPIView):
    model: Type[BaseModel] = None
    serializer_class: Type[ModelSerializer] = None
    search_class: Type[DjangoSearch] = DjangoSearch
    pagination_class = VelaitPagination

    def __init__(self, *args, **kwargs):
        super(SearchView, self).__init__(*args, **kwargs)

        if self.model is None or self.serializer_class is None:
            raise NotImplementedError("Model or Serializer were not supplied to the SearchView")

    def get_search_object(self):
        return self.search_class(
            search_=self.request.GET.get('search'),
            query=self.request.GET.get('query'),
            ordering=self.request.GET.get('ordering'),
            model=self.model,
        )

    @swagger_auto_schema(
        manual_parameters=[
            openapi.Parameter(
                name="search",
                in_=openapi.IN_QUERY,
                type='string',
                description='Search string',
            ),
            openapi.Parameter(
                name="query",
                in_=openapi.IN_QUERY,
                type='[{ fn: string, op: string, fv: any }]',
                description='Query parameter for search',
            ),
            openapi.Parameter(
                name="ordering",
                in_=openapi.IN_QUERY,
                type='string',
                description='Ordering fields separated by commas. Use - in front of the field to order by desc',
            ),
            openapi.Parameter(
                name="page",
                in_=openapi.IN_QUERY,
                type='{ offset: int, page: int }',
                description='Page parameter for pagination.',
            ),
        ],
        pagination_class=VelaitPagination,
        paginator_inspectors=[VelaitPaginationInspector],
    )
    def get(self, request, *args, **kwargs):
        try:
            search = self.get_search_object()
            paginator = self.pagination_class()
            queryset = paginator.paginate_queryset(queryset=search.search(), request=request, view=self)
            return paginator.get_paginated_response(self.serializer_class(instance=queryset, many=True).data)

        except SearchError as exc:
            return APIResponse(errors=[exc], status=400)


class ModelRefsView(GenericAPIView):
    model: Type[BaseModel]
    serializer_class: Type[BaseSerializer]
    model_objects: str = 'objects'
    UUID_FIELD_NAME: str = 'id'
    pagination_class = VelaitPagination

    @swagger_auto_schema(
        manual_parameters=[
            openapi.Parameter(
                name="ids",
                in_=openapi.IN_QUERY,
                type='list',
                description="Ids of the models you need to query",
            ),
        ],
        pagination_class=VelaitPagination,
        paginator_inspectors=[VelaitPaginationInspector],
    )
    def get(self, request: Request, *args, **kwargs):
        searched_ids = request.query_params.get('ids')

        if searched_ids is None:
            return APIResponse(errors={
                VelaitError(
                    name="query",
                    description=_("You must provide 'ids' query parameter to use this view"),
                ),
            })
        elif len(searched_ids) > getattr(settings, 'ID_LISTING_MAX_LEN', 40000):
            return APIResponse(errors={
                VelaitError(
                    name="query",
                    description=_("Your 'ids' are too long. Try shortening them"),
                ),
            })

        numerical_ids, uuids = [], []

        for id_ in searched_ids.split(","):
            if re.fullmatch(r"([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})", id_):
                uuids.append(id_)
            elif re.fullmatch(r'\d+', id_):
                numerical_ids.append(int(id_))
            else:
                return APIResponse(errors={
                    VelaitError(
                        name="query",
                        description=_("Id must be numbers or UUID. Your format is invalid"),
                    ),
                })

        return APIResponse(
            data=self.serializer_class(
                instance=filter_objects(
                    getattr(self.model, self.model_objects),
                    Q(pk__in=numerical_ids) | Q(**{f"{self.UUID_FIELD_NAME}__in": uuids}),
                    limit=getattr(settings, "ID_LISTING_MAX_RESULTS", 300),
                ),
                many=True,
            ).data,
            status=200,
        )


__all__ = [
    'ModelRefsView',
    'SearchView',
]



from rest_framework.exceptions import NotFound
from django.utils.translation import gettext_lazy as _

from velait.common.exceptions import AlreadyDeletedError
from velait.velait_django.main.services.services import delete_object


class SoftDeleteViewMixin:
    def perform_destroy(self, instance):
        try:
            delete_object(instance)
        except AlreadyDeletedError:
            raise NotFound(detail=_("Уже удален"), code="obj")



import hashlib
import random
import string
from time import time

from django.conf import settings


def create_random_str(length: int):
    seed = f"{random.getstate()}{time()}{settings.SECRET_KEY}"
    random.seed(hashlib.sha256(seed.encode()).digest())
    return "".join(random.choice(string.printable) for _ in range(length))


__all__ = ['create_random_str']



from rest_framework import serializers
from rest_framework.utils import model_meta

from velait.velait_django.main.models import BaseModel
from velait.velait_django.main.services.services import update_object


class BaseSerializer(serializers.ModelSerializer):
    id = serializers.UUIDField(source='uuid', read_only=True)

    class Meta:
        model = BaseModel
        fields = (
            "id",
            "is_deleted",
            "created_at",
            "created_by_id",
            "updated_at",
            "updated_by_id",
        )
        read_only_fields = (
            "id",
            "is_deleted",
            "created_at",
            "created_by_id",
            "updated_at",
            "updated_by_id",
        )
        abstract = True

    def get_user_id(self, instance):
        raise NotImplementedError("You need to add get_user() function to your serializers")

    def update(self, instance, validated_data):
        info = model_meta.get_field_info(instance)

        many_to_many_fields = []
        update_values = {}

        for attr, value in validated_data.items():
            if attr in info.relations and info.relations[attr].to_many:
                many_to_many_fields.append((attr, value))
            else:
                update_values[attr] = value

        update_object(model=instance, updated_by_id=self.get_user_id(instance), **update_values)

        for attr, value in many_to_many_fields:
            field = getattr(instance, attr)
            field.set(value)

        return instance

    def create(self, validated_data):
        info = model_meta.get_field_info(self.Meta.model)

        many_to_many = {}

        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and (field_name in validated_data):
                many_to_many[field_name] = validated_data.pop(field_name)

        instance = self.Meta.model.objects.create(**validated_data, created_by_id=self.get_user_id(None))

        if many_to_many:
            for field_name, value in many_to_many.items():
                field = getattr(instance, field_name)
                field.set(value)

        return instance


__all__ = ['BaseSerializer']



from rest_framework.views import exception_handler
from django.core.exceptions import ValidationError
from django.conf import settings

from velait.common.exceptions import VelaitError
from velait.velait_django.main.api.responses import APIResponse


def custom_exception_handler(exc, context):
    response = exception_handler(exc, context)
    response_exists = response is not None

    if not response_exists:
        if isinstance(exc, ValidationError):
            errors = [
                VelaitError(name=error.code, description="\n".join(error.messages))
                for error in exc.error_list
            ]
        else:
            if settings.DEBUG:
                raise exc

            errors = [
                VelaitError(
                    name=getattr(exc, 'name', "Unknown error"),
                    description=getattr(exc, 'description', "Unknown error"),
                )
            ]

        response = APIResponse(data=None, status=400, )
    else:
        if isinstance(response.data, dict):
            errors = [
                VelaitError(name=error_name, description=error_description)
                for error_name, error_description in response.data.items()
            ]
        else:
            errors = response.data

    return APIResponse(
        data=None,
        status=response.status_code,
        template_name=response.template_name,
        headers=response.headers,
        exception=response.exception,
        content_type=response.content_type,
        errors=errors,
    )



from velait.common.exceptions import VelaitError


class SearchError(VelaitError):
    status_code = 400

    def __init__(self, name: str, description: str):
        super(SearchError, self).__init__(description)
        self.description = description
        self.name = name


class PaginationLimitsError(VelaitError):
    description = "page"
    name = "Pagination limits are invalid"
    status_code = 400


__all__ = [
    'SearchError',
    'PaginationLimitsError',
]



from rest_framework.authentication import BaseAuthentication, TokenAuthentication


class MIOSSOAuthentication(BaseAuthentication):
    def authenticate(self, request):
        pass

    @staticmethod
    def get_authorization_header(request):
        pass



from typing import Iterable

from fastapi.responses import JSONResponse

from velait.velait_fastapi.api.schemas import BaseSchema


class APIResponse(JSONResponse):
    def __init__(
        self,
        content=None,
        status_code=200,
        headers=None,
        errors: Iterable['ResponseErrorItem'] = None,
        **kwargs
    ):
        add_pagination = False

        if errors is not None:
            content = {
                "pagination": {
                    "total_records": None,
                    "total_pages": None,
                    "first": None,
                    "last": None,
                    "next": None,
                    "previous": None,
                },
                "results": None,
                "errors": [error.dict() for error in errors],
            }
        elif isinstance(content, dict):
            add_pagination = content.get('pagination') is None
        else:
            add_pagination = True

        if add_pagination:
            content = {
                "pagination": {
                    "total_records": 1,
                    "total_pages": 1,
                    "first": None,
                    "last": None,
                    "next": None,
                    "previous": None,
                },
                "results": content,
                "errors": [],
            }

        super(APIResponse, self).__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            **kwargs,
        )


class ResponseErrorItem(BaseSchema):
    name: str
    description: str


__all__ = [
    'APIResponse',
    'ResponseErrorItem',
]



from django.conf import settings
from django.utils.translation import gettext_lazy as _

from velait.common.services.search import SearchOperator, Search, SearchError
from velait.velait_django.main.services.services import filter_objects


class DjangoSearch(Search):
    DEFAULT_PAGE_SIZE = settings.REST_FRAMEWORK.get('PAGE_SIZE')

    def search(self):
        try:
            return filter_objects(
                objects=self.model.objects,
                ordering=self._parse_ordering(),
                **self.parse_query_filters(),
            )
        except SearchError:
            raise SearchError(
                name="search",
                description=_("Поиск не может быть выполнен"),
            )

    def parse_query_filters(self):
        if self._query is None:
            return {}

        parsed_operators = (self._parse_operator(
            field_name=query_part.get('fn'),
            operator=query_part.get('op'),
            field_value=query_part.get('fv'),
        ) for query_part in self._query)

        return {field: field_filter for field, field_filter in parsed_operators}

    def _parse_operator(self, field_name: str, operator: str, field_value: str):
        """
        Creates an expression object if all input operators are valid.
        If they are not, raises op an exception
        """

        if operator == SearchOperator.LESS.value:
            return [f"{field_name}__lt", field_value]
        elif operator == SearchOperator.EQUAL.value:
            return [field_name, field_value]
        elif operator == SearchOperator.GREATER.value:
            return [f"{field_name}__gt", field_value]
        elif operator == SearchOperator.CONTAINS.value:
            return [f"{field_name}__in", field_value]
        elif operator == SearchOperator.LESS_OR_EQUAL.value:
            return [f"{field_name}__lte", field_value]
        elif operator == SearchOperator.GREATER_OR_EQUAL.value:
            return [f"{field_name}__gte", field_value]
        else:
            raise SearchError(name="query", description=_("Операция не найдена"))


__all__ = [
    'SearchOperator',
    'SearchError',
    'DjangoSearch',
]



from rest_framework.response import Response


class APIResponse(Response):
    @staticmethod
    def format_data(data):
        return {
            "results": [data] if isinstance(data, dict) else data,
            "pagination": {
                'totalRecords': 1,
                'totalPages': 1,
                'first': None,
                'last': None,
                'next': None,
                'previous': None,
            },
            "errors": [],
        }

    def __init__(
        self,
        data=None,
        status=None,
        template_name=None,
        headers=None,
        exception=False,
        content_type=None,
        errors=None,
    ):
        if errors is not None:
            data = {
                "errors": [
                    {'name': error.name, 'description': error.description}
                    for error in errors
                ],
            }

        elif isinstance(data, dict):
            if data.get('pagination') is None:
                data = APIResponse.format_data(data)

        super(APIResponse, self).__init__(
            data=data,
            status=status,
            template_name=template_name,
            headers=headers,
            exception=exception,
            content_type=content_type,
        )



from typing import Iterable

from fastapi import Request

from velait.velait_fastapi.api.users.exceptions import NoPermissionError


class UserGroups:
    BUSINESS_ADMINS = "BusinessAdmins"
    BUSINESS_USERS = "BusinessUsers"
    REQUEST_MANAGERS = "OrgRequestsManagers"
    REQUEST_READERS = "OrgRequestsReaders"
    PLATFORM_ADMINS = "PlatformAdmins"
    PLATFORM_REQUEST_MANAGER = "PlatformRequestsManagers"
    PLATFORM_REQUESTS_READERS = "PlatformRequestsReaders"


class PermissionChecker:
    def __init__(self, permissions: Iterable[str]):
        self.permissions = set(permissions)

    def __call__(self, request: Request = None):
        if not set(request.user.permissions) & set(self.permissions):
            raise NoPermissionError()


__all__ = ['NoPermissionError', 'UserGroups', 'PermissionChecker']



import json
from django.conf import settings
from drf_yasg import openapi
from drf_yasg.inspectors import PaginatorInspector
from rest_framework.utils.urls import replace_query_param, remove_query_param
from rest_framework.pagination import PageNumberPagination

from velait.velait_django.main.api.responses import APIResponse


class VelaitPagination(PageNumberPagination):
    offset: int = None
    page_size: int = None

    def __parse_page_param(self, request):
        try:
            parsed_page = json.loads(request.query_params['page'])

            if not isinstance(parsed_page, dict):
                raise TypeError()

            if parsed_page.get('size') <= 0:
                parsed_page['size'] = settings.REST_FRAMEWORK,

            if parsed_page.get('offset', 0) <= 0:
                parsed_page['offset'] = 1

            self.page_size, self.offset = parsed_page['size'], parsed_page['offset']

        except (json.JSONDecodeError, TypeError, ValueError, KeyError) as exc:
            self.page_size, self.offset = settings.REST_FRAMEWORK['PAGE_SIZE'], 1

    def paginate_queryset(self, queryset, request, view=None):
        self.__parse_page_param(request)
        return super(VelaitPagination, self).paginate_queryset(queryset, request, view)

    def get_next_link(self):
        if not self.page.has_next():
            return None

        return replace_query_param(
            url=self.request.build_absolute_uri(),
            key=self.page_query_param,
            val=json.dumps({ "size": self.page_size, "offset": self.offset + 1 })
        )

    def get_previous_link(self):
        if not self.page.has_previous():
            return None

        url = self.request.build_absolute_uri()
        page_number = self.page.previous_page_number()

        if page_number == 1:
            return remove_query_param(url, self.page_query_param)

        return replace_query_param(
            url=url,
            key=self.page_query_param,
            val=json.dumps({ "size": self.page_size, "offset": self.offset - 1 })
        )

    def get_page_size(self, request):
        return self.page_size

    def get_page_number(self, request, paginator):
        return self.offset

    def get_paginated_response(self, data):
        return APIResponse({
            "pagination": {
                'totalRecords': self.page.paginator.count,
                'totalPages': self.page.paginator.num_pages,
                'first': remove_query_param(
                    url=self.request.build_absolute_uri(),
                    key=self.page_query_param,
                ),
                'last': replace_query_param(
                    url=self.request.build_absolute_uri(),
                    key=self.page_query_param,
                    val=json.dumps({
                        "size": self.page_size,
                        "offset": self.page.paginator.num_pages,
                    }),
                ),
                'next': self.get_next_link(),
                'previous': self.get_previous_link(),
            },
            "results": data,
            "errors": [],
        })


class VelaitPaginationInspector(PaginatorInspector):
    def get_paginated_response(self, paginator, response_schema):
        return openapi.Schema(
            type=openapi.TYPE_OBJECT,
            properties={
                "pagination": openapi.Schema(
                    type=openapi.TYPE_OBJECT,
                    properties={
                        'totalRecords': openapi.Schema(type=openapi.TYPE_INTEGER),
                        'totalPages': openapi.Schema(type=openapi.TYPE_INTEGER),
                        'first': openapi.Schema(type=openapi.TYPE_STRING,
                                                format=openapi.FORMAT_URI, x_nullable=True),
                        'last': openapi.Schema(type=openapi.TYPE_STRING,
                                               format=openapi.FORMAT_URI, x_nullable=True),
                        'next': openapi.Schema(type=openapi.TYPE_STRING,
                                               format=openapi.FORMAT_URI, x_nullable=True),
                        'previous': openapi.Schema(type=openapi.TYPE_STRING,
                                                   format=openapi.FORMAT_URI, x_nullable=True),
                    },
                ),
                'results': response_schema,
                'errors': openapi.Schema(
                    type=openapi.TYPE_ARRAY,
                    items=openapi.Schema(
                        type=openapi.TYPE_OBJECT,
                        properties={
                            'name': openapi.Schema(type=openapi.TYPE_STRING),
                            'description': openapi.Schema(type=openapi.TYPE_STRING),
                        },
                    ),
                ),
            },
            required=['results'],
        )



from django.utils.translation import gettext_lazy as _

from velait.common.exceptions import VelaitError


class UserAuthenticationError(VelaitError):
    name = "auth"
    description = _('Ошибка аутентификации')
    status_code = 400
